# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (HEART) Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a JATIC compatible ART PyTorchClassifier.
"""
from typing import Any, Sequence

from art.estimators.classification.pytorch import PyTorchClassifier
from maite.protocols import ArrayLike

from heart_library.utils import process_inputs_for_art


class JaticPyTorchClassifier(PyTorchClassifier):  # pylint: disable=R0901,R0901
    """
    JATIC compatible extension of ART core PyTorchClassifier
    """

    def __init__(self, provider: str = "", **kwargs: Any):

        if provider == "huggingface":
            import transformers  # pylint: disable=C0415
            from art.estimators.classification.hugging_face import \
                HuggingFaceClassifierPyTorch  # pylint: disable=C0415

            model = transformers.AutoModelForImageClassification.from_pretrained(kwargs["model"])
            kwargs["model"] = model
            hf_model = HuggingFaceClassifierPyTorch(**kwargs)
            kwargs["model"] = hf_model.model
        elif provider == "timm":
            from art.estimators.classification.hugging_face import \
                HuggingFaceClassifierPyTorch  # pylint: disable=C0415
            from timm import create_model  # pylint: disable=C0415

            model = create_model(
                kwargs["model"],
                pretrained=True,
                num_classes=kwargs["nb_classes"],
            )
            kwargs["model"] = model
            hf_model = HuggingFaceClassifierPyTorch(**kwargs)
            kwargs["model"] = hf_model.model

        super().__init__(**kwargs)

    def __call__(self, data: Sequence[ArrayLike]) -> Sequence[ArrayLike]:

        # convert to ART supported type
        images, _, _ = process_inputs_for_art(data)

        # make prediction
        output = self.predict(images)

        return [output]



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (HEART) Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a JATIC compatible ART Attack.
"""
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from art.attacks import Attack, EvasionAttack
from maite.protocols import ArrayLike

from heart_library.estimators.object_detection.pytorch import \
    JaticPyTorchObjectDetectionOutput
from heart_library.utils import process_inputs_for_art


class JaticEvasionAttackOutput:
    """
    Dataclass output JaticEvasionAttackOutput
    """

    def __init__(self, images: List[np.ndarray], targets: np.ndarray, metadata: List[Dict[str, Any]]):
        self.images = images
        self.targets = targets
        self.metadata = metadata

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, ind: int) -> Tuple[np.ndarray, np.ndarray, Dict[str, Any]]:
        return self.images[ind], self.targets[ind], self.metadata[ind]


@dataclass
class JaticPoisonAttackOutput:
    """
    Dataclass output JaticEvasionAttackOutput
    """

    poisoning_examples: ArrayLike
    poisoning_labels: ArrayLike


class JaticAttack:  # pylint: disable=R0901
    """
    Wrapper for JATIC compatible attacks
    """

    def __init__(self, attack: Attack, norm: int = 0):
        self._attack = attack
        self._norm = norm

    def __call__(
        self,
        data: Union[Tuple[Sequence[ArrayLike], Sequence[ArrayLike], Sequence[dict[str, Any]]], ArrayLike],
        **kwargs
    ) -> Tuple[
        Sequence[ArrayLike],
        Union[Sequence[ArrayLike], Sequence[JaticPyTorchObjectDetectionOutput], Optional[Any]],
        Sequence[dict[str, Any]],
    ]:
        """
        Run the attack.. depending on the attack, call different func.. depending on the attack, different output
        .. metrics?
        .. include an argument which would add "benign predictions" to the metadata.. to be used in metrics
        .. where both benign and adversarial images needed.. also store the original image in metadata?
        """

        attack_output: Tuple[
            Sequence[ArrayLike],
            Union[Sequence[ArrayLike], Sequence[JaticPyTorchObjectDetectionOutput], Optional[Any]],
            Sequence[dict[str, Any]],
        ]

        # convert to ART supported data types
        # assume data is a batch of data and returns a modified version of that batch
        x, y, meta = process_inputs_for_art(data)
        is_object_detection = False
        if y is not None and not isinstance(y, np.ndarray):
            # check if y is a list of dicts, as per object detection
            if not all(isinstance(i, dict) for i in y):
                y = np.array(y)
            else:
                is_object_detection = True

        # run the attack
        if isinstance(self._attack, EvasionAttack):

            adv_output = self._attack.generate(x, y, **kwargs)

            # check if adversarial patch attack
            # requries extra step of applying the patch
            if hasattr(self._attack, "apply_patch"):
                from art.attacks.evasion import (  # pylint: disable=C0415
                    DPatch, RobustDPatch)

                if isinstance(self._attack, (RobustDPatch, DPatch)):
                    patch = adv_output
                    mask = np.array([])
                    adv_images = self._attack.apply_patch(x=x)
                else:
                    patch, mask = adv_output
                    adv_images = self._attack.apply_patch(x=x, scale=1)

                if meta is not None and None not in meta and len(meta) > 0:
                    for item in meta:
                        item.update({"patch": patch, "mask": mask})
                else:
                    meta = []
                    for _ in range(len(adv_images)):
                        meta.append({"patch": patch, "mask": mask})

                if is_object_detection:
                    if y is not None:
                        targets = [JaticPyTorchObjectDetectionOutput(detection) for detection in y]
                    else:
                        targets = []
                    attack_output = ([np.asarray(list(adv_images))], targets, meta)
                else:
                    attack_output = [np.asarray(list(adv_images))], y, meta

            # not a patch attack
            else:
                # convert to JATIC supported data types
                if is_object_detection:
                    if y is not None:
                        targets = [JaticPyTorchObjectDetectionOutput(detection) for detection in y]
                    else:
                        targets = []
                    attack_output = ([np.asarray(list(adv_output))], targets, meta)
                else:
                    attack_output = [np.asarray(list(adv_output))], y, meta

            # check if should calculate the difference between benign
            # and adversarial images to store in meta
            if self._norm > 0:
                diff = np.linalg.norm((attack_output[0] - x).reshape(len(x), -1), ord=self._norm, axis=1)
                meta = attack_output[2]
                if meta is not None and None not in meta and len(meta) > 0:
                    for i, item in enumerate(meta):
                        item.update({"delta": diff[i]})
                else:
                    meta = []
                    for item in diff:
                        meta.append({"delta": item})
                    attack_output = (attack_output[0], attack_output[1], meta)

        return attack_output



# MIT License
#
# Copyright (C) HEART Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a HEART compatible ART Query Efficient Black Box attack
"""
from typing import Optional, Union

import numpy as np
from art.attacks import EvasionAttack
from art.attacks.evasion import FastGradientMethod
from art.estimators import BaseEstimator, LossGradientsMixin

from heart_library.estimators.classification import HeartQueryEfficientGradientEstimationClassifier


class HeartQueryEfficientBlackBoxAttack(EvasionAttack):
    """
    HEART defined extension of ART core Query Efficient Black Box attack.
    """

    attack_params = EvasionAttack.attack_params + [
        "norm",
        "eps",
        "eps_step",
        "targeted",
        "num_random_init",
        "batch_size",
        "minimal",
        "summary_writer",
    ]
    _estimator_requirements = (BaseEstimator, LossGradientsMixin)

    def __init__(
        self,
        estimator,
        num_basis: int = 20,
        sigma: float = 1 / 64.0,
        round_samples: float = 0.0,
        norm: Union[int, float, str] = np.inf,
        eps: Union[int, float, np.ndarray] = 0.3,
        eps_step: Union[int, float, np.ndarray] = 0.1,
        targeted: bool = False,
        num_random_init: int = 0,
        batch_size: int = 32,
        minimal: bool = False,
        **kwargs
    ):
        super().__init__(estimator=estimator, **kwargs)

        query_efficient_estimator = HeartQueryEfficientGradientEstimationClassifier(
            estimator, num_basis, sigma, round_samples
        )
        self._attack = FastGradientMethod(
            query_efficient_estimator,
            norm=norm,
            eps=eps,
            eps_step=eps_step,
            targeted=targeted,
            num_random_init=num_random_init,
            batch_size=batch_size,
            minimal=minimal,
        )

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        return self._attack.generate(x)



# MIT License
#
# Copyright (C) HEART Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module extends ART's `LaserAttack` attack to support HEART.

| Paper link: https://arxiv.org/abs/2103.06504
"""

import logging
from typing import Callable, Optional, Tuple, Union

import numpy as np
from art.attacks.evasion.laser_attack.laser_attack import (AdvObjectGenerator,
                                                           LaserAttack,
                                                           LaserBeam,
                                                           LaserBeamGenerator)
from art.attacks.evasion.laser_attack.utils import (AdversarialObject,
                                                    DebugInfo, ImageGenerator)
from art.summary_writer import SummaryWriter

logger = logging.getLogger(__name__)


def greedy_search(
    image: np.ndarray,
    estimator,
    iterations: int,
    actual_class: int,
    actual_class_confidence: float,
    adv_object_generator: AdvObjectGenerator,
    image_generator: ImageGenerator,
    debug: Optional[DebugInfo] = None,
) -> Tuple[Optional[AdversarialObject], Optional[int]]:  # pragma: no cover
    """
    Extending ART's greedy search algorithm to support HEART. Specifically supports channel first images.

    :param image: Image to attack.
    :param estimator: Predictor of the image class.
    :param iterations: Maximum number of iterations of the algorithm.
    :param actual_class:
    :param actual_class_confidence:
    :param adv_object_generator: Object responsible for adversarial object generation.
    :param image_generator: Object responsible for image generation.
    :param debug: Optional debug handler.
    """
    params = adv_object_generator.random()
    for _ in range(iterations):
        for sign in [-1, 1]:
            params_prim = adv_object_generator.update_params(params, sign=sign)
            adversarial_image = image_generator.update_image(image, params_prim)

            # channels first
            adversarial_image = adversarial_image.transpose(0, 3, 1, 2).astype(np.float32)

            prediction = estimator.predict(adversarial_image)
            if debug is not None:
                DebugInfo.report(debug, params_prim, np.squeeze(adversarial_image, 0))
            predicted_class = prediction.argmax()
            confidence_adv = prediction[0][actual_class]
            if confidence_adv <= actual_class_confidence:
                params = params_prim
                actual_class_confidence = confidence_adv
                break

        if predicted_class != actual_class:
            return params, predicted_class
    return None, None


class HeartLaserAttack(LaserAttack):
    """
    Extension of ART's implementation of a generic laser attack case which
    supports channel first images.
    """

    def __init__(
        self,
        estimator,
        iterations: int,
        laser_generator: AdvObjectGenerator,
        image_generator: ImageGenerator = ImageGenerator(),
        random_initializations: int = 1,
        optimisation_algorithm: Callable = greedy_search,
        debug: Optional[DebugInfo] = None,
    ) -> None:
        """
        :param estimator: Predictor of the image class.
        :param iterations: Maximum number of iterations of the algorithm.
        :param laser_generator: Object responsible for generation laser beams images and their update.
        :param image_generator: Object responsible for image generation.
        :param random_initializations: How many times repeat the attack.
        :param optimisation_algorithm: Algorithm used to generate adversarial example. May be replaced.
        :param debug: Optional debug handler.
        """
        super().__init__(
            estimator=estimator,
            iterations=iterations,
            laser_generator=laser_generator,
            image_generator=image_generator,
            random_initializations=random_initializations,
            optimisation_algorithm=optimisation_algorithm,
            debug=debug,
        )

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial examples.

        :param x: Images to attack as a tensor in NHWC order
        :param y: Array of correct classes
        :return: Array of adversarial images
        """

        if x.ndim != 4:  # pragma: no cover
            raise ValueError("Unrecognized input dimension. Only tensors NHWC are acceptable.")

        # channels first
        x = x.transpose(0, 2, 3, 1)

        parameters = self.generate_parameters(x, y)
        adversarial_images = np.zeros_like(x)
        for image_index in range(x.shape[0]):
            laser_params, _ = parameters[image_index]
            if laser_params is None:
                adversarial_images[image_index] = x[image_index]
                continue
            adversarial_image = self._image_generator.update_image(x[image_index], laser_params)
            adversarial_images[image_index] = adversarial_image

        adversarial_images = adversarial_images.transpose(0, 3, 1, 2)
        return adversarial_images

    def _generate_params_for_single_input(
        self, x: np.ndarray, y: Optional[int] = None
    ) -> Tuple[Optional[AdversarialObject], Optional[int]]:  # pragma: no cover
        """
        Generate adversarial example params for a single image.

        :param x: Image to attack as a tensor (NRGB = (1, ...))
        :param y: Correct class of the image. If not provided, it is set to the prediction of the model.
        :return: Adversarial object params and adversarial class number.
        """

        image = np.expand_dims(x, 0)

        # channels first
        prediction = self.estimator.predict(image.transpose(0, 3, 1, 2))

        if y is not None:
            actual_class = y
        else:
            actual_class = prediction.argmax()
        actual_class_confidence = prediction[0][actual_class]

        for _ in range(self.random_initializations):
            laser_params, predicted_class = self._attack_single_image(image, actual_class, actual_class_confidence)
            if laser_params is not None:
                logger.info("Found adversarial params: %s", laser_params)
                return laser_params, predicted_class
        logger.warning("Couldn't find adversarial laser parameters")

        return None, None

    def _check_params(self) -> None:  # pragma: no cover
        """
        Remove restriction against channel first images.
        """
        if not isinstance(self._summary_writer_arg, (bool, str, SummaryWriter)):
            raise ValueError("The argument `summary_writer` has to be either of type bool or str.")
        if self.iterations <= 0:
            raise ValueError("The iterations number has to be positive.")
        if self.random_initializations <= 0:
            raise ValueError("The random initializations has to be positive.")


class HeartLaserBeamAttack(HeartLaserAttack):
    """
    Extension of ART's implementation of the `LaserBeam` attack, which
    supports channel first images.

    | Paper link: https://arxiv.org/abs/2103.06504
    """

    def __init__(
        self,
        estimator,
        iterations: int,
        max_laser_beam: Union[LaserBeam, Tuple[float, float, float, int]],
        min_laser_beam: Union[LaserBeam, Tuple[float, float, float, int]] = (380.0, 0.0, 1.0, 1),
        random_initializations: int = 1,
        image_generator: ImageGenerator = ImageGenerator(),
        debug: Optional[DebugInfo] = None,
    ) -> None:  # pragma: no cover
        """
        :param estimator: Predictor of the image class.
        :param iterations: Maximum number of iterations of the algorithm.
        :param max_laser_beam: LaserBeam with maximal parameters or tuple (wavelength, angle::radians, bias, width)
            of the laser parameters.
        :param min_laser_beam: LaserBeam with minimal parameters or tuple (wavelength, angle::radians, bias, width)
            of the laser parameters.
        :param image_generator: Object responsible for image generation.
        :param random_initializations: How many times repeat the attack.
        :param debug: Optional debug handler.
        """

        if isinstance(min_laser_beam, tuple):
            min_laser_beam_obj = LaserBeam.from_array(list(min_laser_beam))
        else:
            min_laser_beam_obj = min_laser_beam
        if isinstance(max_laser_beam, tuple):
            max_laser_beam_obj = LaserBeam.from_array(list(max_laser_beam))
        else:
            max_laser_beam_obj = max_laser_beam

        super().__init__(
            estimator,
            iterations,
            LaserBeamGenerator(min_laser_beam_obj, max_laser_beam_obj),
            image_generator=image_generator,
            random_initializations=random_initializations,
            debug=debug,
        )



"""
Module providing evasion attacks
"""

from heart_library.attacks.evasion.hop_skip_jump import HeartHopSkipJump
from heart_library.attacks.evasion.laser_attack import HeartLaserBeamAttack
from heart_library.attacks.evasion.query_efficient_bb_attack import \
    HeartQueryEfficientBlackBoxAttack

__all__ = (
    "HeartQueryEfficientBlackBoxAttack",
    "HeartLaserBeamAttack",
    "HeartHopSkipJump",
)



# MIT License
#
# Copyright (C) HEART Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a HEART compatible ART QueryEfficientGradientEstimationClassifier.
"""
import numpy as np
from art.estimators.classification import QueryEfficientGradientEstimationClassifier
from art.utils import clip_and_round


class HeartQueryEfficientGradientEstimationClassifier(
    QueryEfficientGradientEstimationClassifier
):  # pylint: disable=R0901
    """
    HEART compatible extension of ART core QueryEfficientGradientEstimationClassifier
    """

    def __init__(
        self,
        classifier,
        num_basis: int = 20,
        sigma: float = 1 / 64.0,
        round_samples: float = 0.0,
    ):
        super().__init__(classifier=classifier, num_basis=num_basis, sigma=sigma, round_samples=round_samples)

    def predict(self, x: np.ndarray, batch_size: int = 128, **kwargs) -> np.ndarray:  # pylint: disable=W0221
        """
        Perform prediction of the classifier for input `x`. Rounds results first.

        :param x: Features in array of shape (nb_samples, nb_features) or (nb_samples, nb_pixels_1, nb_pixels_2,
                  nb_channels) or (nb_samples, nb_channels, nb_pixels_1, nb_pixels_2).
        :param batch_size: Size of batches.
        :return: Array of predictions of shape `(nb_inputs, nb_classes)`.
        """
        x = clip_and_round(x, self.clip_values, self.round_samples).astype(np.float32)
        return self._classifier.predict(x, batch_size=batch_size)



"""
Module providing classification estimators
"""
from heart_library.estimators.classification.query_efficient_bb import HeartQueryEfficientGradientEstimationClassifier

__all__ = ("HeartQueryEfficientGradientEstimationClassifier",)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (HEART) Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Module implementing varying metrics for assessing model robustness. These fall mainly under two categories:
attack-dependent and attack-independent.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
from typing import Any, Dict, Sequence, Union

import numpy as np

from heart_library.attacks.attack import JaticAttack
from heart_library.estimators.object_detection import \
    JaticPyTorchObjectDetectionOutput

logger = logging.getLogger(__name__)


class HeartMAPMetric:
    """
    Facilitating support for Torchmetric's MAP metric.
    """

    def __init__(self, **kwargs):
        """
        :param **kwargs: arguments passed to Torchmetric's MAP metric
        """
        from torchmetrics.detection.mean_ap import \
            MeanAveragePrecision  # pylint: disable=C0415

        self.metric = MeanAveragePrecision(**kwargs)

    def reset(self) -> None:
        """
        Clear contents of current metric's cache of predictions and targets.
        """
        return self.metric.reset()

    def update(
        self,
        preds_batch: Sequence[JaticPyTorchObjectDetectionOutput],
        targets_batch: Sequence[JaticPyTorchObjectDetectionOutput],
    ) -> None:
        """
        Add predictions and targets to metric's cache for later calculation.
        :param preds_batch: predictions in ObjectDetectionTarget format.
        :param targets_batch: groundtruth targets in ObjectDetectionTarget format.
        """
        import torch  # pylint: disable=C0415

        # Torchmetrics mAP expects list of dicts with one dict per image; each dict with:
        # - boxes: Tensor w/shape (num_boxes, 4)
        # - scores: Tensor w/shape (num_boxes)
        # - labels: Tensor w/shape (num_boxes)
        # iterate over images in batch
        for preds, targets in zip(preds_batch, targets_batch):
            # put predictions and labels in dictionaries
            # tensor bridge to PyTorch tensors (required by Torchmetrics)
            preds_dict = {
                "boxes": torch.as_tensor(preds.boxes),
                "scores": torch.as_tensor(preds.scores),
                "labels": torch.as_tensor(preds.labels),
            }
            targets_dict = {
                "boxes": torch.as_tensor(targets.boxes),
                "scores": torch.as_tensor(targets.scores),
                "labels": torch.as_tensor(targets.labels),
            }
            self.metric.update([preds_dict], [targets_dict])

    def compute(self) -> Dict[str, Any]:
        """
        Compute MAP scores.
        """
        return self.metric.compute()


class HeartAccuracyMetric:
    """
    Facilitating support for Torchmetric's Accuracy metric.
    """

    def __init__(self, is_logits: bool = True, **kwargs):
        """
        :param is_logits: bool indicating if predictions are logits
        :param **kwargs: arguments passed to Torchmetric's Accuracy metric
        """
        from torchmetrics.classification import (  # pylint: disable=C0415
            BinaryAccuracy, MulticlassAccuracy, MultilabelAccuracy)

        self.is_logits = is_logits
        self._metric: Union[BinaryAccuracy, MulticlassAccuracy, MultilabelAccuracy]
        self._task = kwargs.pop("task")
        if self._task == "binary":
            self._metric = BinaryAccuracy(**kwargs)
        elif self._task == "multiclass":
            self._metric = MulticlassAccuracy(**kwargs)
        elif self._task == "multilabel":
            self._metric = MultilabelAccuracy(**kwargs)

    def reset(self) -> None:
        """
        Clear contents of current metric's cache of predictions and targets.
        """
        return self._metric.reset()

    def update(self, preds_batch: Sequence[np.ndarray], targets_batch: Sequence[np.ndarray]) -> None:
        """
        Add predictions and targets to metric's cache for later calculation.
        :param preds_batch: predictions in numpy array format
        :param targets_batch: groundtruth targets in numpy array format
        """
        import torch  # pylint: disable=C0415

        if self.is_logits:
            preds = torch.as_tensor(np.argmax(preds_batch, axis=2).ravel())
        else:
            preds = torch.as_tensor(preds_batch)
        targets = torch.as_tensor(targets_batch)
        self._metric.update(preds, targets)

    def compute(self) -> Dict[str, float]:
        """
        Compute accuracy score.
        """
        return {"accuracy": self._metric.compute().item()}


class RobustnessBiasMetric:
    """
    A metric which describes Robustness Bias for features
    of datasets. Currently supports only classification tasks.
    """

    def __init__(self, metadata: Sequence[Dict[str, Any]], labels: np.ndarray, interval: int = 100):
        """
        :param metadata: Sequence[Dict[str, Any]] - the metadata computed during attack
            which contains delta between benign and adversarial images
        :param labels: List[str] - classification labels
        :param interval: int - tau
        """
        self._state: Dict = {}
        self._labels: np.ndarray = labels
        self._metadata: Sequence[Dict[str, Any]] = metadata
        self._interval: int = interval

    def reset(self):
        """
        Reset the metric to default values.
        """
        self._state = {}

    def update(self, preds_batch: Sequence[np.ndarray], targets_batch: Sequence[np.ndarray]) -> None:
        """
        Add predictions and targets to metric's cache for later calculation.
        :param preds_batch: predictions in numpy array format
        :param targets_batch: groundtruth targets in numpy array format
        """
        try:
            errors = np.stack([item["delta"] for item in self._metadata])
        except KeyError as key_error:
            raise KeyError(
                "Delta not computed for metadata. Set norm > 0 of JaticAttack to compute delta."
            ) from key_error

        # assuming the targets batch is the groundtruth or original predictions
        # and the preds batch are predictions for the augmented / attacked data
        success = (np.argmax(targets_batch, axis=2).ravel() != np.argmax(preds_batch, axis=2).ravel()).astype(int)

        taus = np.linspace(0, max(errors) + 1, self._interval)
        series: Dict = {}
        for tau in taus:
            for label in range(len(self._labels)):
                idxs_of_label = np.argwhere(np.argmax(targets_batch, axis=2).ravel() == label).ravel()
                idxs_of_label_success = np.argwhere(success[idxs_of_label] == 1).ravel()
                idxs_of_label = idxs_of_label[idxs_of_label_success]
                errors_of_label = errors[idxs_of_label]
                idx_error_greater_tau = np.argwhere(errors_of_label > tau).ravel()
                if len(errors_of_label) != 0:
                    proportion = len(idx_error_greater_tau) / len(errors_of_label)
                    if label in series:
                        series[label].append([tau, proportion])
                    else:
                        series[label] = [[tau, proportion]]

        self._state = series

    def compute(self) -> Dict:
        """
        Returns the computed metric
        """
        return self._state


class AccuracyPerturbationMetric:
    """
    A metric for easily calculating the clean and robust accuracy
    as well as the perturbation between clean and adversarial input.
    """

    def __init__(
        self,
        benign_predictions: Sequence[np.ndarray],
        metadata: Sequence[Dict[str, Any]],
        accuracy_type: str = "robust",
    ):
        """
        :param accuracy_type: str - the type of accuracy to calculate. Choice of "adversarial" or "robust".
            - Robust accuracy is the accuracy of the model on all samples
            - Adversarial accuracy is the accuracy of the model only samples which were
              correctly predicted in the non-adversarial scenario
        """
        self._state: Dict = {}
        self._benign_predictions = benign_predictions
        self._metadata = metadata
        self._accuracy_type = accuracy_type

    def reset(self):
        """
        Reset the metric to default values.
        """
        self._state = {}

    def update(self, preds_batch: Sequence[np.ndarray], targets_batch: Sequence[np.ndarray]):
        """
        Updates the metric value.
        """
        y_orig = np.argmax(self._benign_predictions, axis=2).ravel()
        y_pred = np.argmax(preds_batch, axis=2).ravel()

        try:
            mean_delta = np.stack([item["delta"] for item in self._metadata]).mean()
        except KeyError as key_error:
            raise KeyError(
                "Delta not computed for metadata. Set norm > 0 of JaticAttack to compute delta."
            ) from key_error

        y_corr = y_orig == targets_batch

        clean_acc = np.sum(y_corr) / len(y_orig)
        attack_acc: float = 0.0
        if self._accuracy_type == "adversarial":
            attack_acc = np.sum((y_pred == y_orig) & y_corr) / np.sum(y_corr)
        elif self._accuracy_type == "robust":
            attack_acc = np.mean(y_pred == targets_batch)

        self._state = {
            "clean_accuracy": clean_acc,
            f"{self._accuracy_type}_accuracy": attack_acc,
            "mean_delta": mean_delta,
        }

    def compute(self) -> Dict[str, float]:
        """
        Returns the computed metric
        in Tuple (clean_accuracy, robust_accuracy, average_perturbation)
        """
        return self._state


class BlackBoxAttackQualityMetric:
    """
    A metric for extracting the black box quality metrics.
    """

    def __init__(self, attack: JaticAttack):
        """
        :param attack: JaticAttack - the black-box attack (currently only HopSkipJump supported)
        """
        self._state: Dict = {}
        self._attack = attack._attack

    def reset(self):
        """
        Reset the metric to default values.
        """
        self._state = {}

    def update(self):
        """
        Updates the metric value.
        """
        total_queries = self._attack.total_queries
        adv_query_idx = self._attack.adv_query_idx
        adv_queries = [len(item) for item in adv_query_idx]
        benign_queries = [total_queries[i] - n_adv for i, n_adv in enumerate(adv_queries)]
        adv_perturb_total = self._attack.perturbs
        adv_perturb_iter = self._attack.perturbs_iter
        adv_confs_total = self._attack.confs

        self._state = {
            "total_queries": total_queries,
            "adv_queries": adv_queries,
            "benign_queries": benign_queries,
            "adv_query_idx": adv_query_idx,
            "adv_perturb_total": adv_perturb_total,
            "adv_perturb_iter": adv_perturb_iter,
            "adv_confs_total": adv_confs_total,
        }

    def compute(self) -> Dict[str, Any]:
        """
        Returns the computed metric
        in dict
        """
        return self._state



# MIT License
#
# Copyright (C) HEART Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a JATIC compatible ART Object Detector.
"""
import sys
from typing import Any, Dict, List, Sequence, Union

import numpy as np
import torch
from art.estimators.object_detection import (PyTorchDetectionTransformer,
                                             PyTorchFasterRCNN,
                                             PyTorchObjectDetector,
                                             PyTorchYolo)
from maite.protocols import ArrayLike
from torchvision.models import detection as fasterrcnn

from heart_library.utils import process_inputs_for_art

if sys.version_info >= (3, 10):
    import yolov5
    from yolov5.utils.loss import ComputeLoss

    class Yolo(torch.nn.Module):
        """
        Wrapper for YOLO object detection models
        """

        def __init__(self, model, device):
            super().__init__()
            self.model = model
            self.model.hyp = {
                "box": 0.05,
                "obj": 1.0,
                "cls": 0.5,
                "anchor_t": 4.0,
                "cls_pw": 1.0,
                "obj_pw": 1.0,
                "fl_gamma": 0.0,
            }
            self.model.model.model.to(device)  # Required when using GPU

            self.compute_loss = ComputeLoss(self.model.model.model)

        def forward(self, x, targets=None):
            """
            Modified forward to facilitate computation of loss dict
            """
            if self.training:
                outputs = self.model.model.model(x)
                loss, loss_items = self.compute_loss(outputs, targets)
                loss_components_dict = {"loss_total": loss}
                loss_components_dict["loss_box"] = loss_items[0]
                loss_components_dict["loss_obj"] = loss_items[1]
                loss_components_dict["loss_cls"] = loss_items[2]
                return loss_components_dict
            return self.model(x)


COCO_YOLO_LABELS = [
    "person",
    "bicycle",
    "car",
    "motorcycle",
    "airplane",
    "bus",
    "train",
    "truck",
    "boat",
    "traffic light",
    "fire hydrant",
    "stop sign",
    "parking meter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "backpack",
    "umbrella",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sports ball",
    "kite",
    "baseball bat",
    "baseball glove",
    "skateboard",
    "surfboard",
    "tennis racket",
    "bottle",
    "wine glass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hot dog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "couch",
    "potted plant",
    "bed",
    "dining table",
    "toilet",
    "tv",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cell phone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddy bear",
    "hair drier",
    "toothbrush",
]
COCO_FASTER_RCNN_LABELS = COCO_DETR_LABELS = [
    "background",
    "person",
    "bicycle",
    "car",
    "motorbike",
    "aeroplane",
    "bus",
    "train",
    "truck",
    "boat",
    "trafficlight",
    "firehydrant",
    "streetsign",
    "stopsign",
    "parkingmeter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "hat",
    "backpack",
    "umbrella",
    "shoe",
    "eyeglasses",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sportsball",
    "kite",
    "baseballbat",
    "baseballglove",
    "skateboard",
    "surfboard",
    "tennisracket",
    "bottle",
    "plate",
    "wineglass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hotdog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "sofa",
    "pottedplant",
    "bed",
    "mirror",
    "diningtable",
    "window",
    "desk",
    "toilet",
    "door",
    "tvmonitor",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cellphone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "blender",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddybear",
    "hairdrier",
    "toothbrush",
    "hairbrush",
]
SUPPORTED_DETECTORS: dict = {
    "yolov5s": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5n": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5m": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5l": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5x": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5n6": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5s6": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5m6": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5l6": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "yolov5x6": "YOLOv5 model. Ref: https://github.com/ultralytics/yolov5",
    "fasterrcnn_resnet50_fpn": "Faster R-CNN model. Ref: \
https://pytorch.org/vision/master/models/generated/torchvision.models\
.detection.fasterrcnn_resnet50_fpn.html#\
torchvision.models.detection.fasterrcnn_resnet50_fpn",
    "fasterrcnn_resnet50_fpn_v2": "Faster R-CNN model. Ref: \
https://pytorch.org/vision/master/models/generated/torchvision.models.\
detection.fasterrcnn_resnet50_fpn_v2.html#\
torchvision.models.detection.fasterrcnn_resnet50_fpn_v2",
    "fasterrcnn_mobilenet_v3_large_fpn": "Faster R-CNN model. Ref: \
https://pytorch.org/vision/master/models/generated/torchvision.models.\
detection.fasterrcnn_mobilenet_v3_large_fpn.html#\
torchvision.models.detection.fasterrcnn_mobilenet_v3_large_fpn",
    "fasterrcnn_mobilenet_v3_large_320_fpn": "Faster R-CNN model. Ref: \
https://pytorch.org/vision/master/models/generated/torchvision.models\
.detection.fasterrcnn_mobilenet_v3_large_320_fpn.html#\
torchvision.models.detection.fasterrcnn_mobilenet_v3_large_320_fpn",
    "detr_resnet50": "Detection Transformer. Ref: https://github.com/facebookresearch/detr",
    "detr_resnet101": "Detection Transformer. Ref: https://github.com/facebookresearch/detr",
    "detr_resnet50_dc5": "Detection Transformer. Ref: https://github.com/facebookresearch/detr",
    "detr_resnet101_dc5": "Detection Transformer. Ref: https://github.com/facebookresearch/detr",
}


class JaticPyTorchObjectDetectionOutput:
    """
    Object Detection Output
    """

    def __init__(self, detection: Dict[str, np.ndarray]):
        """
        param: Dict[str, np.ndarray] - detection data
        """
        self._boxes = detection["boxes"]
        self._labels = detection["labels"]
        self._scores = detection["scores"]

    @property
    def boxes(self) -> np.ndarray:
        """
        Return detection bounding boxes
        """
        return self._boxes

    @boxes.setter
    def boxes(self, value):
        """
        Update detection bounding boxes
        """
        self._boxes = value

    @property
    def labels(self) -> np.ndarray:
        """
        Return detection labels
        """
        return self._labels

    @labels.setter
    def labels(self, value):
        """
        Update detection labels
        """
        self._labels = value

    @property
    def scores(self) -> np.ndarray:
        """
        Return detection scores
        """
        return self._scores

    @scores.setter
    def scores(self, value):
        """
        Update detection scores
        """
        self._scores = value


class JaticPyTorchObjectDetector(PyTorchObjectDetector):  # pylint: disable=R0901
    """
    JATIC compatible extension of ART core PyTorchObjectDetector
    """

    def __init__(self, model: Union["torch.nn.Module", str] = "", model_type: str = "", **kwargs: Any):
        """
        :param model: Union[torch.nn.Module, str] - a loaded model or path to model
        :param type: str - one of supported_detectors e.g. yolov5.
        """
        self.model_type = model_type

        if "models" in sys.modules:
            sys.modules.pop("models")

        if isinstance(model, torch.nn.Module):
            super().__init__(model=model, **kwargs)
            if "yolo" in model_type:
                self._detector = PyTorchYolo(model, **kwargs)
            elif "detr" in model_type:
                self._detector = PyTorchDetectionTransformer(model, **kwargs)
            elif "fasterrcnn" in model_type:
                self._detector = PyTorchFasterRCNN(model, **kwargs)
            else:
                raise ValueError(f"Model type {model_type} is not supported. Try one of {SUPPORTED_DETECTORS}.")

        elif isinstance(model, str):

            if "device_type" in kwargs:
                device_type = kwargs["device_type"]
            else:
                if not torch.cuda.is_available():
                    device_type = torch.device("cpu")
                else:  # pragma: no cover
                    cuda_idx = torch.cuda.current_device()
                    device_type = torch.device(f"cuda:{cuda_idx}")

            if model_type == "":
                raise ValueError(
                    f"To use a local model, please specify param: type, \
                        with one of the supported models: {SUPPORTED_DETECTORS}"
                )

            # YOLO
            if "yolo" in model_type:
                if sys.version_info >= (3, 10):
                    try:
                        if model == "":
                            loaded_model = Yolo(yolov5.load(f"{model_type}.pt"), device_type)
                        else:  # pragma: no cover
                            loaded_model = Yolo(yolov5.load(model), device_type)
                    except Exception as load_exception:
                        raise Exception("Yolov5 model was not successful loaded.") from load_exception
                    self._detector = PyTorchYolo(loaded_model, **kwargs)
                else:
                    raise ValueError("yolov5 models require python versions 3.10 and above.")

            # DETR
            elif "detr" in model_type:
                if model == "":
                    loaded_model = torch.hub.load("facebookresearch/detr", model_type, pretrained=True)
                else:  # pragma: no cover
                    checkpoint = torch.load(model, map_location=device_type)
                    loaded_model = torch.hub.load("facebookresearch/detr", model_type, pretrained=False)
                    loaded_model.load_state_dict(checkpoint["model"])
                self._detector = PyTorchDetectionTransformer(loaded_model, **kwargs)

            # Faster-RCNN
            elif "fasterrcnn" in model_type:
                if model == "":
                    frcnn_detector = getattr(fasterrcnn, model_type)
                    loaded_model = frcnn_detector(
                        pretrained=True, progress=True, num_classes=91, pretrained_backbone=True
                    )
                else:  # pragma: no cover
                    checkpoint = torch.load(model, map_location=device_type)
                    n_classes = checkpoint["roi_heads.box_predictor.cls_score.weight"].shape[0]
                    frcnn_detector = getattr(fasterrcnn, model_type)
                    loaded_model = frcnn_detector(
                        pretrained=False, progress=True, num_classes=n_classes, pretrained_backbone=True
                    )
                    loaded_model.load_state_dict(checkpoint)
                self._detector = PyTorchFasterRCNN(loaded_model, **kwargs)
            else:
                raise ValueError(f"Model type {model_type} is not supported. Try one of {SUPPORTED_DETECTORS}.")

    def __getattr__(self, attr):
        return getattr(self._detector, attr)

    def __call__(self, data: ArrayLike) -> Sequence[JaticPyTorchObjectDetectionOutput]:
        # convert to ART supported type
        images, _, _ = process_inputs_for_art(data)

        # make prediction
        output = self._detector.predict(images)

        # convert back to JATIC supported type
        return [JaticPyTorchObjectDetectionOutput(detection) for detection in output]

    def _translate_labels(self, labels: List[Dict[str, "torch.Tensor"]]) -> Any:
        """Route to method of instantiated detector"""
        return self._detector._translate_labels(labels)  # pylint: disable=W0212

    def _translate_predictions(self, predictions: Any) -> List[Dict[str, np.ndarray]]:
        """Route to method of instantiated detector"""
        return self._detector._translate_predictions(predictions)  # pylint: disable=W0212



"""
The HEART extension.
"""
import logging.config

# Project Imports

# Semantic Version
__version__ = "0.4.3"

# pylint: disable=C0103

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "std": {
            "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M",
        }
    },
    "handlers": {
        "default": {
            "class": "logging.NullHandler",
        },
        "test": {
            "class": "logging.StreamHandler",
            "formatter": "std",
            "level": logging.INFO,
        },
    },
    "loggers": {
        "heart": {"handlers": ["default"]},
        "tests": {"handlers": ["test"], "level": "INFO", "propagate": True},
    },
}
logging.config.dictConfig(LOGGING)
logger = logging.getLogger(__name__)



# MIT License
#
# Copyright (C) HEART Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module extends ART's `HopSkipJump` attack to support HEART.
"""

from typing import List, Optional, Tuple, Union

import numpy as np
from art.attacks.evasion import HopSkipJump
from art.config import ART_NUMPY_DTYPE
from art.utils import (check_and_transform_label_format, get_labels_np_array,
                       to_categorical)
from tqdm.auto import tqdm


def softmax(x):  # pragma: no cover
    """Compute softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum()


class HeartHopSkipJump(HopSkipJump):
    """
    Extension of ART's implementation of a generic laser attack case which
    supports channel first images.
    """

    def __init__(
        self,
        classifier,
        batch_size: int = 64,
        targeted: bool = False,
        norm: Union[int, float, str] = 2,
        max_iter: int = 50,
        max_eval: int = 10000,
        init_eval: int = 100,
        init_size: int = 100,
        verbose: bool = True,
    ) -> None:
        """
        Create a HopSkipJump attack instance.

        :param classifier: A trained classifier.
        :param batch_size: The size of the batch used by the estimator during inference.
        :param targeted: Should the attack target one specific class.
        :param norm: Order of the norm. Possible values: "inf", np.inf or 2.
        :param max_iter: Maximum number of iterations.
        :param max_eval: Maximum number of evaluations for estimating gradient.
        :param init_eval: Initial number of evaluations for estimating gradient.
        :param init_size: Maximum number of trials for initial generation of adversarial examples.
        :param verbose: Show progress bars.
        """
        super().__init__(
            classifier=classifier,
            batch_size=batch_size,
            targeted=targeted,
            norm=norm,
            max_iter=max_iter,
            max_eval=max_eval,
            init_eval=init_eval,
            init_size=init_size,
            verbose=verbose,
        )

        self.total_queries: np.ndarray = np.array([])
        self.adv_query_idx: List = []
        self.perturbs: List = []
        self.perturbs_iter: List = []
        self.confs: List = []
        self.curr_idx: int = 0
        self.curr_val: np.ndarray = np.array([])

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:  # pragma: no cover
        """
        Generate adversarial samples and return them in an array.

        :param x: An array with the original inputs to be attacked.
        :param y: Target values (class labels) one-hot-encoded of shape `(nb_samples, nb_classes)` or indices of shape
                  (nb_samples,).
        :param mask: An array with a mask broadcastable to input `x` defining where to apply adversarial perturbations.
                     Shape needs to be broadcastable to the shape of x and can also be of the same shape as `x`. Any
                     features for which the mask is zero will not be adversarially perturbed.
        :type mask: `np.ndarray`
        :param x_adv_init: Initial array to act as initial adversarial examples. Same shape as `x`.
        :type x_adv_init: `np.ndarray`
        :param resume: Allow users to continue their previous attack.
        :type resume: `bool`
        :return: An array holding the adversarial examples.
        """
        mask = kwargs.get("mask")
        self.total_queries = np.zeros(x.shape[0])
        self.adv_query_idx = [[] for i in range(x.shape[0])]
        self.perturbs = [[] for i in range(x.shape[0])]
        self.perturbs_iter = [[] for i in range(x.shape[0])]
        self.confs = [[] for i in range(x.shape[0])]
        self.curr_idx = 0
        self.curr_val = np.array([])

        if y is None:
            # Throw error if attack is targeted, but no targets are provided
            if self.targeted:  # pragma: no cover
                raise ValueError("Target labels `y` need to be provided for a targeted attack.")

            # Use model predictions as correct outputs
            y = get_labels_np_array(self.estimator.predict(x, batch_size=self.batch_size))  # type: ignore
            self.total_queries = self.total_queries + 1

        y = check_and_transform_label_format(y, nb_classes=self.estimator.nb_classes)
        if not isinstance(y, np.ndarray):
            raise ValueError("Targets not correctly formatted.")
        if self.estimator.nb_classes == 2 and y.shape[1] == 1:  # pragma: no cover
            raise ValueError(
                "This attack has not yet been tested for binary classification with a single output classifier."
            )

        # Check whether users need a stateful attack
        resume = kwargs.get("resume")

        if resume is not None and resume:
            start = self.curr_iter
        else:
            start = 0

        # Check the mask
        if mask is not None:
            if len(mask.shape) == len(x.shape):
                mask = mask.astype(ART_NUMPY_DTYPE)
            else:
                mask = np.array([mask.astype(ART_NUMPY_DTYPE)] * x.shape[0])
        else:
            mask = np.array([None] * x.shape[0])

        # Get clip_min and clip_max from the classifier or infer them from data
        if self.estimator.clip_values is not None:
            clip_min, clip_max = self.estimator.clip_values
        else:
            clip_min, clip_max = np.min(x), np.max(x)

        # Prediction from the original images
        preds = np.argmax(self.estimator.predict(x, batch_size=self.batch_size), axis=1)
        self.total_queries = self.total_queries + 1

        # Prediction from the initial adversarial examples if not None
        x_adv_init = kwargs.get("x_adv_init")

        if x_adv_init is not None:
            # Add mask param to the x_adv_init
            for i in range(x.shape[0]):
                if mask[i] is not None:
                    x_adv_init[i] = x_adv_init[i] * mask[i] + x[i] * (1 - mask[i])

            # Do prediction on the init
            init_pred_logits = self.estimator.predict(x_adv_init, batch_size=self.batch_size)
            init_preds = np.argmax(init_pred_logits, axis=1)
            self.total_queries = self.total_queries + 1
            for i in range(x_adv_init.shape[0]):
                dist = np.linalg.norm(x[i] - x_adv_init[i])
                self.perturbs[i] += [dist]
                # self.confs[i] += [np.max(softmax(init_pred_logits[i]))]

        else:
            init_preds = [None] * len(x)
            x_adv_init = [None] * len(x)

        # Assert that, if attack is targeted, y is provided
        if self.targeted and y is None:  # pragma: no cover
            raise ValueError("Target labels `y` need to be provided for a targeted attack.")

        # Some initial setups
        x_adv = x.astype(ART_NUMPY_DTYPE)
        self.original_image = x

        y = np.argmax(y, axis=1)

        # Generate the adversarial samples
        for ind, val in enumerate(tqdm(x_adv, desc="HopSkipJump", disable=not self.verbose)):
            self.curr_iter: int = start
            self.curr_idx = ind
            self.curr_val = val

            if self.targeted:
                x_adv[ind] = self._perturb(
                    x=val,
                    y=y[ind],  # type: ignore
                    y_p=preds[ind],
                    init_pred=init_preds[ind],
                    adv_init=x_adv_init[ind],
                    mask=mask[ind],
                    clip_min=clip_min,
                    clip_max=clip_max,
                )

            else:
                x_adv[ind] = self._perturb(
                    x=val,
                    y=-1,
                    y_p=preds[ind],
                    init_pred=init_preds[ind],
                    adv_init=x_adv_init[ind],
                    mask=mask[ind],
                    clip_min=clip_min,
                    clip_max=clip_max,
                )

        y = to_categorical(y, self.estimator.nb_classes)  # type: ignore

        return x_adv

    def _init_sample(
        self,
        x: np.ndarray,
        y: int,
        y_p: int,
        init_pred: int,
        adv_init: np.ndarray,
        mask: Optional[np.ndarray],
        clip_min: float,
        clip_max: float,
    ) -> Optional[Union[np.ndarray, Tuple[np.ndarray, int]]]:  # pragma: no cover
        """
        Find initial adversarial example for the attack.

        :param x: An array with 1 original input to be attacked.
        :param y: If `self.targeted` is true, then `y` represents the target label.
        :param y_p: The predicted label of x.
        :param init_pred: The predicted label of the initial image.
        :param adv_init: Initial array to act as an initial adversarial example.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :return: An adversarial example.
        """
        nprd = np.random.RandomState()
        initial_sample = None

        if self.targeted:
            # Attack satisfied
            if y == y_p:
                return None

            # Attack unsatisfied yet and the initial image satisfied
            if adv_init is not None and init_pred == y:
                return adv_init.astype(ART_NUMPY_DTYPE), init_pred

            # Attack unsatisfied yet and the initial image unsatisfied
            for _ in range(self.init_size):
                random_img = nprd.uniform(clip_min, clip_max, size=x.shape).astype(x.dtype)

                if mask is not None:
                    random_img = random_img * mask + x * (1 - mask)

                pred_logits = self.estimator.predict(np.array([random_img]), batch_size=self.batch_size)
                random_class = np.argmax(
                    pred_logits,
                    axis=1,
                )[0]
                self.total_queries[self.curr_idx] += np.array([random_img]).shape[0]
                dist = np.linalg.norm(x - random_img)
                self.perturbs[self.curr_idx] += [dist]

                if random_class == y:
                    self.adv_query_idx[self.curr_idx].append(self.total_queries[self.curr_idx])
                    self.confs[self.curr_idx] += [np.max(softmax(pred_logits[0]))]

                    # Binary search to reduce the l2 distance to the original image
                    random_img = self._binary_search(
                        current_sample=random_img,
                        original_sample=x,
                        target=y,
                        norm=2,
                        clip_min=clip_min,
                        clip_max=clip_max,
                        threshold=0.001,
                    )
                    initial_sample = random_img, random_class

                    break
        else:
            # The initial image satisfied
            if adv_init is not None and init_pred != y_p:
                return adv_init.astype(ART_NUMPY_DTYPE), y_p

            # The initial image unsatisfied
            for _ in range(self.init_size):
                random_img = nprd.uniform(clip_min, clip_max, size=x.shape).astype(x.dtype)

                if mask is not None:
                    random_img = random_img * mask + x * (1 - mask)

                pred_logits = self.estimator.predict(np.array([random_img]), batch_size=self.batch_size)
                random_class = np.argmax(
                    pred_logits,
                    axis=1,
                )[0]
                self.total_queries[self.curr_idx] += np.array([random_img]).shape[0]
                dist = np.linalg.norm(x - random_img)
                self.perturbs[self.curr_idx] += [dist]

                if random_class != y_p:
                    self.adv_query_idx[self.curr_idx].append(self.total_queries[self.curr_idx])
                    self.confs[self.curr_idx] += [np.max(softmax(pred_logits[0]))]

                    # Binary search to reduce the l2 distance to the original image
                    random_img = self._binary_search(
                        current_sample=random_img,
                        original_sample=x,
                        target=y_p,
                        norm=2,
                        clip_min=clip_min,
                        clip_max=clip_max,
                        threshold=0.001,
                    )
                    initial_sample = random_img, y_p
                    break

        return initial_sample

    def _attack(
        self,
        initial_sample: np.ndarray,
        original_sample: np.ndarray,
        target: int,
        mask: Optional[np.ndarray],
        clip_min: float,
        clip_max: float,
    ) -> np.ndarray:  # pragma: no cover
        """
        Main function for the boundary attack.

        :param initial_sample: An initial adversarial example.
        :param original_sample: The original input.
        :param target: The target label.
        :param mask: An array with a mask to be applied to the adversarial perturbations. Shape needs to be
                     broadcastable to the shape of x. Any features for which the mask is zero will not be adversarially
                     perturbed.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :return: an adversarial example.
        """
        # Set current perturbed image to the initial image
        current_sample = initial_sample
        # print(f'Image {self.curr_idx} starting eval at query idx {self.total_queries[self.curr_idx]}')
        # Main loop to wander around the boundary
        for _ in range(self.max_iter):
            # First compute delta
            delta = self._compute_delta(
                current_sample=current_sample,
                original_sample=original_sample,
                clip_min=clip_min,
                clip_max=clip_max,
            )

            # Then run binary search
            current_sample = self._binary_search(
                current_sample=current_sample,
                original_sample=original_sample,
                norm=self.norm,
                target=target,
                clip_min=clip_min,
                clip_max=clip_max,
            )

            # Next compute the number of evaluations and compute the update
            num_eval = min(int(self.init_eval * np.sqrt(self.curr_iter + 1)), self.max_eval)

            update = self._compute_update(
                current_sample=current_sample,
                num_eval=num_eval,
                delta=delta,
                target=target,
                mask=mask,
                clip_min=clip_min,
                clip_max=clip_max,
            )

            # Finally run step size search by first computing epsilon
            if self.norm == 2:
                dist = np.linalg.norm(original_sample - current_sample)
            else:
                dist = np.max(abs(original_sample - current_sample))

            epsilon = 2.0 * dist / np.sqrt(self.curr_iter + 1)
            success = False

            while not success:
                epsilon /= 2.0
                potential_sample = current_sample + epsilon * update
                success = self._adversarial_satisfactory(  # type: ignore
                    samples=potential_sample[None],
                    target=target,
                    clip_min=clip_min,
                    clip_max=clip_max,
                )

            # Update current sample
            current_sample = np.clip(potential_sample, clip_min, clip_max)

            # Calc perturbs for iter
            dist = np.linalg.norm(self.original_image[self.curr_idx] - current_sample)
            self.perturbs_iter[self.curr_idx] += [dist]

            # Update current iteration
            self.curr_iter += 1

            # If attack failed. return original sample
            if np.isnan(current_sample).any():  # pragma: no cover
                return original_sample

        return current_sample

    def _adversarial_satisfactory(
        self, samples: np.ndarray, target: int, clip_min: float, clip_max: float
    ) -> np.ndarray:  # pragma: no cover
        """
        Check whether an image is adversarial.

        :param samples: A batch of examples.
        :param target: The target label.
        :param clip_min: Minimum value of an example.
        :param clip_max: Maximum value of an example.
        :return: An array of 0/1.
        """
        samples = np.clip(samples, clip_min, clip_max)
        pred_logits = self.estimator.predict(samples, batch_size=self.batch_size)
        preds = np.argmax(pred_logits, axis=1)
        self.total_queries[self.curr_idx] += samples.shape[0]  # as num_eval might make > 1

        if self.targeted:
            result = preds == target
        else:
            result = preds != target

        for i, res in enumerate(result):
            if res:
                dist = np.linalg.norm(self.curr_val - samples[i])
                self.perturbs[self.curr_idx] += [dist]
                self.confs[self.curr_idx] += [np.max(softmax(pred_logits[i]))]
                self.adv_query_idx[self.curr_idx].append(self.total_queries[self.curr_idx] - len(result) + (i + 1))
        return result



"""
Module providing metrics
"""

from heart_library.metrics.metrics import (
    AccuracyPerturbationMetric,
    BlackBoxAttackQualityMetric,
    HeartAccuracyMetric,
    HeartMAPMetric,
    RobustnessBiasMetric,
)

__all__ = (
    "AccuracyPerturbationMetric",
    "RobustnessBiasMetric",
    "BlackBoxAttackQualityMetric",
    "HeartMAPMetric",
    "HeartAccuracyMetric",
)



# MIT License
#
# Copyright (C) HEART Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Utility methods for converting data types to ART compatible versions.
"""
import logging
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from datasets import Dataset as HFDataset
from datasets.iterable_dataset import IterableDataset as HFIterableDataset
from numpy.typing import NDArray
from PIL.Image import Image as PILImage
from torch import Tensor, is_tensor
from torch.utils.data.dataset import Subset as TorchSubsetDataset

logger = logging.getLogger(__name__)

EMPTY_STRING = ""


class ImageDataset:
    """
    MAITE aligned dataset
    """

    def __init__(self, images: List[np.ndarray], targets: np.ndarray, metadata: List[Dict[str, Any]]):
        self.images = images
        self.targets = targets
        self.metadata = metadata

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, ind: int) -> Tuple[np.ndarray, np.ndarray, Dict[str, Any]]:
        return self.images[ind], self.targets[ind], self.metadata[ind]


def hf_dataset_to_maite(
    dataset: HFDataset,
    image_label: str = EMPTY_STRING,
    target_label: str = EMPTY_STRING,
    meta_label: str = EMPTY_STRING,
    indices: Optional[Sequence[int]] = None,
) -> ImageDataset:
    """
    Convert HF dataset to MAITE aligned dataset
    """
    if indices is None:
        indices = []

    images = []
    target_list = []
    metadata = []

    if len(indices) == 0:
        for item in dataset:

            if image_label is not EMPTY_STRING and image_label in dataset.features:
                images.append(item[image_label])
            elif "img" in dataset.features:
                images.append(item["img"])
            elif "image" in dataset.features:
                images.append(item["image"])
            else:
                raise ValueError("Image feature not found in dataset.")

            if target_label is not EMPTY_STRING:
                target_list.append(item[target_label])
            elif "label" in dataset.features:
                target_list.append(item["label"])

            if meta_label is not EMPTY_STRING:
                metadata.append(item[meta_label])
    else:
        for idx in indices:
            item = dataset[idx]
            if image_label is not EMPTY_STRING and image_label in dataset.features:
                images.append(item[image_label])
            elif "img" in dataset.features:
                images.append(item["img"])
            elif "image" in dataset.features:
                images.append(item["image"])
            else:
                raise ValueError("Image feature not found in dataset.")

            if target_label is not EMPTY_STRING:
                target_list.append(item[target_label])
            elif "label" in dataset.features:
                target_list.append(item["label"])

            if meta_label is not EMPTY_STRING:
                metadata.append(item[meta_label])

    if len(target_list) == 0:
        targets = np.array([None] * len(images))
    else:
        targets = np.array(target_list)

    if len(metadata) == 0:
        metadata = [{} for i in range(len(images))]

    return ImageDataset(images, targets, metadata)


def torch_subset_to_maite(dataset: TorchSubsetDataset) -> ImageDataset:
    """
    Convert Torch subset dataset to MAITE aligned dataset
    """
    images = []
    data = iter(dataset)
    target_list = []
    for item in data:
        images.append(item[0])
        target_list.append(item[1])
    targets = np.array(target_list)
    metadata: List[Dict[str, Any]] = [{} for i in range(len(images))]

    return ImageDataset(images, targets, metadata)


def process_inputs_for_art(
    data: Union[
        HFDataset,
        HFIterableDataset,
        TorchSubsetDataset,
        np.ndarray,
        Tensor,
        Dict,
        Tuple,
        Sequence,
    ]
) -> Tuple[NDArray, Optional[NDArray], List[Dict[str, Any]]]:
    """
    Convert JATIC supported data to ART supported data
    """
    images = np.array([])
    # convert Hugging Face
    if isinstance(data, (HFDataset, HFIterableDataset)):
        data = hf_dataset_to_maite(data)

    # convert Hugging Face if wrapped in Torch Subset
    elif isinstance(data, TorchSubsetDataset) and isinstance(data.dataset, HFDataset):
        data = hf_dataset_to_maite(data.dataset, indices=data.indices)

    elif isinstance(data, TorchSubsetDataset):
        data = torch_subset_to_maite(data)

    # if np.ndarray, convert images to np.ndarray. No targets or metadata.
    if isinstance(data, np.ndarray):
        images = data
        targets = None
        metadata: List[Dict[str, Any]] = []

    # if torch.Tensor, convert images to np.ndarray. No targets or metadata.
    elif is_tensor(data):
        images = np.asarray(data)
        targets = None
        metadata = []

    # if Dict, assume np.ndarray and set values if present.
    elif isinstance(data, Dict) and "images" in data:
        images = data["images"]
        targets = None
        metadata = []

        if "labels" in data:
            targets = data["labels"]

            # for object detection, convert to correct format if exist
            if hasattr(targets[0], "boxes") and hasattr(targets[0], "scores"):
                targets = [
                    {
                        "boxes": np.asarray(t.boxes).astype(np.float32),
                        "scores": np.asarray(t.scores).astype(np.float32),
                        "labels": np.asarray(t.labels).astype(np.int64),
                    }
                    for t in targets
                ]
            elif isinstance(targets[0], np.ndarray):
                targets = np.asarray(targets)
            elif not any(targets):
                targets = None
            else:
                targets = np.asarray(targets)

        if "metadata" in data:
            metadata = data["metadata"]

    # if Tuple of batched data, convert to np.ndarray
    elif isinstance(data, tuple) and isinstance(data[0], (list, np.ndarray, Tensor)):
        images = np.asarray(data[0]).astype(np.float32)
        targets = data[1]
        metadata = data[2]

        # for object detection, convert to correct format if exist
        if hasattr(targets[0], "boxes") and hasattr(targets[0], "scores"):
            targets = [
                {
                    "boxes": np.asarray(t.boxes).astype(np.float32),
                    "scores": np.asarray(t.scores).astype(np.float32),
                    "labels": np.asarray(t.labels).astype(np.int64),
                }
                for t in targets
            ]
        elif isinstance(targets[0], np.ndarray) or is_tensor(targets[0]):
            targets = np.asarray(targets)
        elif not any(targets):
            targets = None
        else:
            targets = np.asarray(targets)

    # if data is a Sequence
    elif isinstance(data, Sequence):
        # here assuming data is a tensor - what if a numpy array?
        # what if each batch (len 1 or >1), has different shape?
        # - in this case will stack fail as different dim images
        # - should auto pad or resize? what if this occurs in other data formats?
        # a sequence of image batches
        if data[0].ndim == 4:
            images = np.vstack([np.asarray(batch) for batch in data])
            targets = None
            metadata = []
        # a sequence of single images
        else:
            images = np.stack([np.asarray(batch) for batch in data])
            targets = None
            metadata = []

    # if Dataset, convert to np.ndarray
    else:
        if hasattr(data, "__getitem__"):
            image_list = []
            targets = []
            metadata = []
            for item in data:
                image_list.append(item[0])
                targets.append(item[1])
                metadata.append(item[2])

        elif hasattr(data, "__iter__"):
            image_list = [item[0] for item in data]
            targets = [item[1] for item in data]
            metadata = [item[2] for item in data if len(item) == 3]

        else:
            raise ValueError("Dataset must implement __getitem__ or __iter__.")

        # check images, targets and meta are same length
        if len(image_list) != len(targets):
            raise ValueError("Images and targets must be same length.")

        # check not empty data
        if not len(image_list) > 0:
            raise ValueError("Images should not be empty.")

        # if images are in PIL format, convert to np.ndarray
        if isinstance(image_list[0], PILImage):
            images = np.asarray(image_list).transpose(0, 3, 1, 2).astype(np.float32)

        # if images are np.ndarray or torch.Tensor, convert to np.ndarray
        # facilitating batches
        elif isinstance(image_list[0], np.ndarray) or is_tensor(image_list[0]):
            if image_list[0].ndim == 3:
                images = np.asarray(image_list).astype(np.float32)
            else:
                images = np.concatenate([np.asarray(batch) for batch in image_list])

        # if images are a list, convert to np.ndarray
        elif isinstance(image_list[0], List):
            images = np.asarray(image_list).astype(np.float32)

        # for object detection, convert to correct format if exist
        if hasattr(targets[0], "boxes") and hasattr(targets[0], "scores"):
            targets = [
                {
                    "boxes": np.asarray(t.boxes).astype(np.float32),
                    "scores": np.asarray(t.scores).astype(np.float32),
                    "labels": np.asarray(t.labels).astype(np.int64),
                }
                for t in targets
            ]
        elif isinstance(targets[0], np.ndarray) or is_tensor(targets[0]):
            targets = np.asarray(targets)
        elif not any(targets):
            targets = None
        else:
            targets = np.asarray(targets)
    return images, targets, metadata



"""
Module providing object detection estimators
"""

from heart_library.estimators.object_detection.pytorch import (
    COCO_DETR_LABELS,
    COCO_FASTER_RCNN_LABELS,
    COCO_YOLO_LABELS,
    SUPPORTED_DETECTORS,
    JaticPyTorchObjectDetectionOutput,
    JaticPyTorchObjectDetector,
)

__all__ = (
    "JaticPyTorchObjectDetector",
    "JaticPyTorchObjectDetectionOutput",
    "SUPPORTED_DETECTORS",
    "COCO_YOLO_LABELS",
    "COCO_FASTER_RCNN_LABELS",
    "COCO_DETR_LABELS",
)
